from doka.kernel.definition.entities import ProjectEntity, UnitEntity
from doka.config import doka_config


class EnvStorage:
    _filename = None

    def read(self) -> dict:
        result = {}
        with open(self._filename, 'rt') as f:
            for line in f:
                line = line.strip()
                if len(line) > 0 and '=' in line:
                    key, value = line.split("=")
                    key = key.strip()
                    value = value.strip()
                    result[key] = value
        return result or None

    def save(self, data: dict):
        with open(self._filename, 'wt') as f:
            for key, value in data.items():
                f.write("{}={}\n".format(key, value))
        return True


class StacksEnvStorage(EnvStorage):
    def __init__(self, unit: UnitEntity):
        self._filename = doka_config.STORAGE_DIR.joinpath(
            *["projects", unit.project_name, "stacks", ".{}.env".format(unit.name)]
        )


class ProjectEnvStorage(EnvStorage):
    def __init__(self, project: ProjectEntity):
        self._filename = project.metadir.joinpath(*["environments", ".env.{}".format(project.env)])


class UnitEnvStorage(EnvStorage):
    def __init__(self, unit: UnitEntity):
        self._filename = unit.codedir.joinpath(".env.example")



from doka.kernel.repository.repositories import ProjectRepository, StackRepository
from doka.kernel.actions import stack


class Stack(object):

    def __init__(self, build: bool = False, branch: str = None, pull: str = None, project=None, join_network=None):
        self._flags = {
            "build": build,
            "branch": branch,
            "pull": pull,
            "join_network": join_network
        }
        if project:
            self._project = ProjectRepository().get_one(project)
        else:
            self._project = ProjectRepository().get_active()

    def build(self, name):
        stack.build(self._stack_entity(name))

    def up(self, name):
        stack_entity = self._stack_entity(name)
        if not stack_entity:
            print("Stack {} not defined".format(name))
            return
        project_entity = ProjectRepository().get_one(stack_entity.project_name)
        if self._flags.get("branch"):
            print("Checkout stack to branch: ", self._flags.get("branch"))
            stack.checkout(stack_entity, branch=self._flags.get("branch"))
        if self._flags.get("pull"):
            print("Pull data from remote origin")
            self._flags.pop("build", None)
            stack.pull(stack_entity)
        if self._flags.get("build"):
            print("Build images")
            stack.build(stack_entity)
        if self._flags.get("join_network"):
            stack.join_network(stack_entity, self._flags.get("join_network"))
        # stack.build(self._stack_entity(name))
        stack.up(stack_entity, project_entity)

    def down(self, name):
        stack.down(self._stack_entity(name))

    def restart(self, name):
        self.down(name)
        self.up(name)

    def _stack_entity(self, stack_name):
        return StackRepository(self._project).get_one(stack_name)



class RepositoryException(Exception):
    _message = None

    def __init__(self):
        super().__init__(self._message)


class EntityExistsException(RepositoryException):
    _message = "Entity exists"


class EntityNotExistsException(RepositoryException):
    _message = "Entity not exists"



import json
from doka.config import doka_config
from doka.kernel.definition.entities import ProjectEntity


class JsonStorage:
    _data = None
    _filename = None

    def read(self):
        if self._data is None:
            try:
                with open(self._filename, 'rt') as f:
                    self._data = json.load(f)
            except:
                return None
        return self._data

    def write(self, data: dict, filename=None):
        try:
            filename = filename or self._filename
            with open(filename, 'wt') as f:
                json.dump(data, f, indent=2)
            self._data = data
            return True
        except Exception as e:
            return False


class ProjectStorage(JsonStorage):
    _filename = doka_config.STORAGE_DIR.joinpath("projects.json")


class UnitStorage(JsonStorage):
    def __init__(self, project: ProjectEntity):
        self._filename = project.metadir.joinpath("units.json")


class StackStorage(JsonStorage):
    def __init__(self, project: ProjectEntity):
        self._filename = project.metadir.joinpath("stacks.json")



from doka.kernel.definition.entities import UnitEntity
from doka.kernel.services.git import Git
from doka.kernel.repository.repositories import UnitRepository, ProjectRepository, StackRepository
from doka.kernel.storage import hooks
from doka.kernel.storage.code import CodeStorage
from doka.kernel.actions import unit, stack
from doka.kernel.services import docker, tester


class Unit(object):

    def __init__(self, template=None, repo=None, stack=None, project=None, branch=None, pull: bool = None,
                 dev: bool = None, ports: str = None, name: str = None, build: bool = None,
                 no_cache: bool = None, mount=None):
        self._flags = {
            "template": template,
            "repo": repo,
            "stack": stack,
            "branch": branch,
            "pull": pull,
            "dev": dev,
            "mount": mount,
            "ports": ports,
            "name": name,
            "build": build,
            "no_cache": no_cache
        }
        if project:
            self._project = ProjectRepository().get_one(project)
        else:
            self._project = ProjectRepository().get_active()

    def bootstrap(self, unit_name):
        template = self._flags.get("template")
        repo = self._flags.get("repo")
        stack = self._flags.get("stack")

        if not template:
            return 'Error: template not defined'
        if not stack:
            return 'Error: stack not defined'

        try:

            unit_entity = UnitEntity(
                name=unit_name,
                project_name=self._project.name,
                repo=template,
                codedir=self._project.workdir.joinpath(*[stack, unit_name]),
                stack_name=stack
            )
            git = Git(unit_entity)
            git.clone(from_template=True)
            git.init()
            if repo:
                unit_entity.repo = repo
                git = Git(unit_entity)
                git.add_origin(push_initial_commit=True)
            git.create_branch("development")

            unit_repository = UnitRepository(self._project)
            unit_repository.save(unit_entity)

        except Exception as e:
            return str(e)

    def build(self, unit_name):
        self._flags["build"] = None
        unit_entity = self._get_unit(unit_name)
        docker.Unit(unit_entity).build(self._flags.get("no_cache"))

    def get(self, repo):
        unit_name = self._flags.get("name")
        if not unit_name:
            unit_name = repo.split('/').pop().replace(".git", "")
        stack_entity = StackRepository(self._project).get_one(self._flags.get("stack") or 'unstacked')
        if not stack_entity:
            return 'Stack {} not defined'.format(stack_entity.name)
        unit_repository = UnitRepository(self._project)
        unit_entity = unit_repository.get_by_any_name(unit_name)

        if not unit_entity:
            unit_entity = UnitEntity(
                name=".".join([stack_entity.name, unit_name]),
                project_name=self._project.name,
                repo=repo,
                codedir=self._project.workdir.joinpath(*[stack_entity.name, unit_name]),
                stack_name=stack_entity.name,
                working_dir='/code'
            )
        git = Git(unit_entity)
        if not unit_entity.codedir.exists():
            git.clone()
            if self._flags.get('branch'):
                git.checkout(self._flags.get("branch"))
        else:
            if self._flags.get('branch'):
                git.checkout(self._flags.get("branch"))
            git.pull()
        try:
            unit_repository.save(unit_entity)
        except:
            pass
        stack.register_unit_in_compose_file(stack_entity, unit_entity)
        hooks.move_unit_environment_to_stack(unit_entity)
        if self._flags.get("build"):
            docker.Unit(unit_entity).build(self._flags.get("no_cache"))

    def test(self, unit_name):
        unit_entity = UnitRepository(self._project).get_by_any_name(unit_name)
        if not unit_entity:
            return 'Unit not found'
        tester.Unit(unit_entity).test_api()

    def up(self, unit_name):
        try:
            unit.start(self._get_unit(unit_name))
        except Exception as e:
            return str(e)

    def down(self, unit_name):
        try:
            unit.stop(self._get_unit(unit_name))
        except Exception as e:
            return str(e)

    def restart(self, unit_name):
        try:
            unit.restart(self._get_unit(unit_name))
        except Exception as e:
            return str(e)

    def scale(self, unit_name, num):
        try:
            unit_entity = UnitRepository(self._project).get_by_any_name(unit_name)
            unit.scale(unit_entity, num)
        except Exception as e:
            return str(e)

    def remove(self, unit_name):
        repository = UnitRepository(self._project)
        unit_entity = repository.get_by_any_name(unit_name)
        if not unit_entity:
            return "Unit not found"
        stack_entity = StackRepository(self._project).get_one(unit_entity.stack_name)
        try:
            unit.stop(unit_entity)
        except Exception as e:
            print("Unit is not running. Skip stopping...")
        repository.remove(unit_entity)
        CodeStorage(unit_entity.codedir).remove()
        stack.unregister_unit_in_compose_file(stack_entity, unit_entity)
        hooks.delete_unit_environment_from_stack(unit_entity)
        return "Unit {} has been removed".format(unit_name)

    def _get_unit(self, unit_name):
        unit_entity = UnitRepository(self._project).get_by_any_name(unit_name)
        if self._flags.get("mount"):
            unit.mount(unit_entity, self._flags.get("mount"))
        if self._flags.get("branch"):
            unit.checkout(unit_entity, self._flags.get("branch"), False)
        if self._flags.get("pull"):
            unit.pull(unit_entity)
        if self._flags.get("build"):
            docker.Unit(unit_entity).build(self._flags.get("no_cache"))
        hooks.move_unit_environment_to_stack(unit_entity)
        return unit_entity



import yaml

dumper = yaml.dumper.SafeDumper
dumper.ignore_aliases = lambda self, data: True
dumper.add_representer(type(None), lambda self, _: self.represent_scalar('tag:yaml.org,2002:null', ''))


class YamlStorage:
    _APPEND_MODE = 'append'
    _SET_MODE = 'set'

    def __init__(self, file):
        self._compose_file = file
        self._data = self.read()
        self._change_node_mode = None

    def set(self, index, value):
        self._change_node_mode = self._SET_MODE
        self._change_node(index, value, self._data)

    def append(self, index, value):
        self._change_node_mode = self._APPEND_MODE
        self._change_node(index, value, self._data)

    def get(self, index):
        data = self._data
        if index:
            for p in index.split('.'):
                data = data.get(p)
                if not data:
                    return None
        return data

    def remove(self, index):
        self._remove_node(index, self._data)

    def read(self):
        with open(self._compose_file, 'rt') as f:
            return yaml.load(f, Loader=yaml.FullLoader)

    def write(self, file=None):
        file = file or self._compose_file
        with open(file, 'wt') as f:
            yaml.dump(self._data, f, default_flow_style=False, sort_keys=False, Dumper=dumper)

    def _remove_node(self, index, node):
        keys = index.split(".")
        key = keys.pop(0)
        if len(keys) == 1:
            node[key] = None
            return
        self._remove_node(".".join(keys), node.get(key))

    def _change_node(self, index, value, node):
        keys = index.split(".")
        key = keys.pop(0)
        if keys:
            if key == '*':
                for _, _node in node.items():
                    self._change_node(".".join(keys), value, _node)
            else:
                self._change_node(".".join(keys), value, node.get(key))
        else:
            if self._change_node_mode == self._APPEND_MODE and isinstance(node.get(key), list):
                if not isinstance(value, list):
                    value = [value]
                node[key] += value
            elif self._change_node_mode == self._APPEND_MODE and \
                    isinstance(node.get(key), dict) and isinstance(value, dict):
                node[key] = {**node[key], **value}
            else:
                node[key] = value



from doka.kernel.definition.entities import UnitEntity
import json
import requests
from doka.kernel.services import docker
import uuid
from termcolor import colored


class Unit:
    def __init__(self, unit_entity: UnitEntity, silent_mode=False):
        self._unit_entity = unit_entity
        self._silent_mode = silent_mode

    def test_api(self):
        try:
            host = docker.Runtime().get_unit_ip(self._unit_entity)
        except:
            self._print(colored("Не могу получить IP сервиса. Возможно он не поднят", "red"))
            return
        for entry in self._unit_entity.codedir.joinpath(*['tests', 'api']).glob('**/*.json'):
            with entry.open('rt') as f:
                scenario = json.load(f)
                self._print("Сценарий:", colored("{}".format(scenario.get("title")), "yellow"))
                for case in scenario.get("cases", []):
                    self._print("\tКейс: ", colored(case.get("title"), "yellow"))
                    self._print("\t\tВыполняем операцию ", colored(case.get("operation"), "yellow"),
                                " с параметрами ",
                                colored(case.get("request"), "yellow"))
                    try:
                        response = requests.post('http://{}/api/jsonrpc'.format(host), json={
                            "jsonrpc": "2.0",
                            "method": case.get("operation"),
                            "id": str(uuid.uuid4()),
                            "params": case.get("request")
                        }).json()
                        self._print("\t\tПолучили ответ: ",
                                    colored(response.get("result") or response.get("error"), "yellow"))
                        self._print("\t\tДолжны были получить: ", colored(case.get("response"), "yellow"))
                        if response.get("result") != case.get("response"):
                            self._print(colored("\t\tОшибка!", "red"))
                        else:
                            self._print(colored("\t\tУспешно!", "green"))
                    except Exception as e:
                        import traceback
                        traceback.print_exc()
                        self._print(colored("\t\tОшибка", "red"), " при отправке запроса: ", str(e))

    def _print(self, *args):
        if not self._silent_mode:
            print(*args)



from pathlib import PosixPath
import shutil


class CodeStorage:
    def __init__(self, directory: PosixPath):
        self._directory = directory

    def remove(self):
        if self._directory.exists():
            shutil.rmtree(self._directory)



from doka.kernel.storage.json import ProjectStorage, UnitStorage, StackStorage
from doka.kernel.definition.entities import ProjectSchema, UnitSchema, StackSchema
from doka.kernel.repository.filters import entity_filter
from doka.kernel.repository.exceptions import *


class Repository:
    _collection = None
    _storage = None
    _schema = None
    _primary_field = None

    def get_one(self, value):
        for entity in self._collection:
            if getattr(entity, self._primary_field) == value:
                return entity
        return None

    def count(self):
        return len(self._collection)

    def get(self, search_filter: dict = None):
        if search_filter is None:
            return self._collection
        return list(filter(entity_filter(search_filter), self._collection))

    def remove(self, entity):
        for i, item in enumerate(self._collection):
            if getattr(item, self._primary_field) == getattr(entity, self._primary_field):
                del self._collection[i]
        self._flush()

    def save(self, entity):
        self._collection = [
            _e for _e in self._collection if
            getattr(_e, self._primary_field) != getattr(entity, self._primary_field)
        ]
        self._collection.append(entity)
        self._flush()

    def _flush(self):
        data = self._schema(many=True).dump(self._collection)
        self._storage.write(data.data)

    def _get_collection(self):
        data = self._storage.read()
        if data:
            return self._schema(many=True).load(data).data
        return []


class UnitRepository(Repository):
    def __init__(self, project):
        if isinstance(project, str):
            project = ProjectRepository().get_one(project)
        self._storage = UnitStorage(project)
        self._schema = UnitSchema
        self._primary_field = "name"
        self._collection = self._get_collection()

    def get_by_any_name(self, name):
        for unit in self._collection:
            if unit.name == name or unit.container_name == name:
                return unit

    def get_for_build(self, stack_name=None):
        result = []
        for unit in self._collection:
            if unit.repo is None:
                continue
            if stack_name and unit.stack_name != stack_name:
                continue
            result.append(unit)
        return result


class ProjectRepository(Repository):
    def __init__(self):
        self._storage = ProjectStorage()
        self._schema = ProjectSchema
        self._primary_field = "name"
        self._collection = self._get_collection()

    def get_active(self):
        for project in self._collection:
            if project.active:
                return project
        return None

    def set_active(self, project_name):
        for project in self._collection:
            project.active = project.name == project_name
        self._flush()


class StackRepository(Repository):
    def __init__(self, project):
        if isinstance(project, str):
            project = ProjectRepository().get_one(project)
        self._storage = StackStorage(project)
        self._schema = StackSchema
        self._primary_field = "name"
        self._collection = self._get_collection()



from marshmallow import Schema, fields, post_load
from pathlib import Path


class ProjectEntity:
    def __init__(self, name, project_repo, deploy_repo, workdir, metadir, env, active=False, docker_registry = None):
        self.name = name
        self.project_repo = project_repo
        self.deploy_repo = deploy_repo
        self.workdir = Path(workdir)
        self.metadir = Path(metadir)
        self.docker_registry = docker_registry
        self.env = env
        self.active = active


class StackEntity:
    def __init__(self, name, project_name, compose_directory, network="default", volumes=None):
        self.name = name
        self.project_name = project_name
        self.compose_directory = Path(compose_directory)
        self.compose_file = "{}.yml".format(self.name)
        self.network = network
        self.volumes = volumes


class UnitEntity:
    def __init__(self, name, project_name, stack_name, image_name=None, repo=None, codedir=None, working_dir=None,
                 branch=None):
        self.project_name = project_name
        self.name = name
        self.image_name = image_name
        self.repo = repo
        self.working_dir = working_dir  # working directory inside container
        self.stack_name = stack_name
        self.branch = branch
        try:
            self.codedir = Path(codedir)  # local code directory
        except:
            self.codedir = codedir

    @property
    def container_name(self):
        return ".".join(self.name.split(".")[1:])


class ProjectSchema(Schema):
    name = fields.Str()
    project_repo = fields.Str()
    deploy_repo = fields.Str()
    workdir = fields.Str()
    metadir = fields.Str()
    env = fields.Str()
    active = fields.Bool()
    docker_registry = fields.Str(missing=None)

    @post_load
    def make(self, data):
        return ProjectEntity(**data)


class StackSchema(Schema):
    name = fields.Str()
    project_name = fields.Str()
    compose_directory = fields.Str()
    network = fields.Str()
    volumes = fields.Dict(missing=None)

    @post_load
    def make(self, data):
        return StackEntity(**data)


class UnitSchema(Schema):
    name = fields.Str()
    image_name = fields.Str(missing=None)
    project_name = fields.Str()
    repo = fields.Str(missing=None)
    codedir = fields.Str(missing=None)
    working_dir = fields.Str(missing=None)
    stack_name = fields.Str()
    branch = fields.Str(missing=None)

    @post_load
    def make(self, data):
        return UnitEntity(**data)
